""" pynchon.api.render

    Basically this is core jinja-rendering stuff.

    Looking for CLI entrypoints?  See pynchon.util.text.render
    Looking for the JinjaPlugin?  See pynchon.plugins.jinja
"""

import os
import functools

from jinja2 import Environment  # Template,; UndefinedError,
from jinja2 import FileSystemLoader, StrictUndefined

from pynchon import abcs, constants, events
from pynchon.util.os import invoke

import jinja2  # noqa

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


def is_templated(txt: str = "") -> bool:
    """ """
    return txt is not None and ("{{" in txt and "}}" in txt)


def dictionary(input, context):
    """
    :param context:
    """
    from pynchon.abcs.visitor import JinjaDict

    return JinjaDict(input).render(context)


@functools.lru_cache(maxsize=None)
def get_jinja_filters():
    def strip_empty_lines(s):
        lines = s.split("\n")
        filtered_lines = [line for line in lines if line.strip() != ""]
        return "\n".join(filtered_lines)

    return dict(
        Path=abcs.Path,
        strip_empty_lines=strip_empty_lines,
    )


@functools.lru_cache(maxsize=None)
def get_jinja_globals():
    """ """
    events.lifecycle.send(__name__, msg="finalizing jinja globals")

    def is_markdown_list_item(text: str):
        return text.rstrip().startswith("* ")

    def invoke_helper(*args, **kwargs) -> typing.StringMaybe:
        """A jinja filter/extension"""
        strict = kwargs.pop("strict", True)
        out = invoke(*args, **kwargs)
        if not out.succeeded:
            raise Exception(out)
        return out.stdout

    def markdown_toc(fname: str, level=None):
        """ """
        import markdown

        with open(fname) as fhandle:
            contents = fhandle.read()
        md = markdown.Markdown(
            extensions=["toc", "fenced_code"],
            extension_configs={"toc": {"toc_depth": level}},
        )
        html = md.convert(contents)
        return md.toc

    # markdown-toc --type github  --no-write docs/blog/ambient-calculus-1.md
    # assert fname
    # fname = abcs.Path(fname)
    # assert fname.exists()
    # # script = abcs.Path(pynchon.__file__).parents[0] / "scripts" / "gh-md-toc.sh"
    # result = invoke(
    #     f"markdown-toc --type github --no-write {fname}",
    #     command_logger=LOGGER.critical
    # )
    # assert result.succeeded
    # return result.stdout
    return dict(
        str=str,
        sh=invoke_helper,
        bash=invoke_helper,
        is_markdown_list_item=is_markdown_list_item,
        invoke=invoke_helper,
        map=map,
        markdown_toc=markdown_toc,
        eval=eval,
        env=os.getenv,
        filter=filter,
    )


def get_jinja_includes(*includes):
    """ """
    includes = list(includes)
    includes += list(constants.PYNCHON_CORE_INCLUDES_DIRS)

    return [abcs.Path(t) for t in includes]


@functools.lru_cache(maxsize=None)
def get_jinja_env(*includes, quiet: bool = False):
    """ """
    events.lifecycle.send(__name__, msg="finalizing jinja-Env")
    includes = get_jinja_includes(*includes)
    for template_dir in includes:
        if not template_dir.exists:
            err = f"template directory @ `{template_dir}` does not exist"
            raise ValueError(err)
    # includes and (not quiet) and LOGGER.warning(f"Includes: {includes}")
    env = Environment(
        loader=FileSystemLoader([str(t) for t in includes]),
        undefined=StrictUndefined,
        # trim_blocks=True,
        # lstrip_blocks=True
    )

    # from jinja2 import Environment, FileSystemLoader
    # Function to include a template
    def include_template(template_name):
        # env = Environment(loader=FileSystemLoader('templates'))  # Assuming templates are in 'templates' directory
        template = env.get_template(template_name)
        return template.render()

    env.filters["include"] = include_template

    env.filters.update(**get_jinja_filters())
    env.pynchon_includes = includes

    env.globals.update(**get_jinja_globals())

    known_templates = list(map(abcs.Path, set(env.loader.list_templates())))

    if known_templates:
        from pynchon.util import text as util_text

        msg = "Known template search paths (includes folders only): "
        tmp = list({p.parents[0] for p in known_templates})
        LOGGER.info(msg + util_text.to_json(tmp))
    return env


def get_template(
    template_name: typing.Union[str, abcs.Path] = None,
    env=None,
    from_string: str = None,
    **jinja_context,
):
    """ """
    env = env or get_jinja_env()
    if isinstance(template_name, (abcs.Path,)):
        template_path = template_name
        template_name = str(template_name)
    elif template_name:
        template_path = abcs.Path(template_name)
    else:
        template_path = None
    try:
        if from_string:
            template = env.from_string(from_string)
        else:
            LOGGER.info(f"Looking up {template_path}")
            template = env.get_template(template_name)
    except (jinja2.exceptions.TemplateNotFound,) as exc:
        LOGGER.critical(f"Template exception: {exc}")
        LOGGER.critical(f"Jinja-includes: {env.pynchon_includes}")
        err = getattr(exc, "templates", exc.message)
        LOGGER.critical(f"Problem template: {err}")
        raise
    jinja_context.update(__template__=template_path)
    all_extensions = filter(
        None, template_path.name[template_path.name.find(".") :].split(".")
    )
    if template_path and "md" in all_extensions:
        LOGGER.warning("template is markdown, trying to parse metadata")
        import markdown

        from pynchon.util.text import loads

        try:
            # https://python-markdown.github.io/extensions/meta_data/#accessing-the-meta-data
            md = markdown.Markdown(extensions=["meta"])
            html = md.convert(open(template_name).read())
            md_meta = md.Meta
            md_meta and LOGGER.warning(f"extracted metadata: {md_meta}")
            # the metadata extension doesn't really do much parsing,
            # so we treat it here as potentially yaml
            tmp = {}
            for k, v in md_meta.items():
                z = "\n".join(v)
                tmp.update(**loads.yaml(f"{k}: {z}"))
            md_meta = tmp
            jinja_context.update(**md_meta)
        except (Exception,) as exc:
            LOGGER.warning(f"failed extracting markdown metadata: {exc}")

    def panic():
        raise Exception(jinja_context)

    def jinja_debug():
        LOGGER.critical(f"jinja_debug: {jinja_context}")
        return str(jinja_context)

    env.globals.update(
        jinja_debug=jinja_debug,
        panic=panic,
    )

    template.render = functools.partial(template.render, **jinja_context)
    return template


def get_template_from_string(content, **kwargs):
    """ """
    return get_template(from_string=content, **kwargs)


def get_template_from_file(
    file: str = None,
    **kwargs,
):
    """
    :param file: str = None:
    :param **kwargs:
    """
    with open(file) as fhandle:
        content = fhandle.read()
    return get_template_from_string(content, file=file, **kwargs)


def clean_whitespace(txt: str):
    # return txt
    return "\n".join([x for x in txt.split("\n") if x.strip()])



""" pynchon.plugins.dockerhub
"""

import webbrowser

from fleks import cli, tagging  # noqa

from pynchon import abcs, events, models  # noqa
from pynchon.util import files, lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Dockerhub(models.Provider):
    """Context for Dockerhub"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "dockerhub"
        org_name: typing.StringMaybe = typing.Field(default=None, help="")
        repo_name: typing.StringMaybe = typing.Field(default=None, help="")

        @property
        def org_url(self) -> str:
            return self.org_name and f"https://hub.docker.com/r/{self.org_name}"

        @property
        def repo_url(self) -> str:
            return (
                self.repo_name and self.org_url and f"{self.org_url}/{self.repo_name}"
            )

    name = "dockerhub"
    cli_name = "dockerhub"

    def open(self):
        """Open this dockerhub project's webpage"""
        url = self.config.repo_url or self.config.org_url
        webbrowser.open(url)



""" pynchon.plugins.plugins
"""

from fleks import tagging

from pynchon import abcs, cli, models
from pynchon.api import render
from pynchon.util.os import invoke
from pynchon.util.text import dumps

from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class PluginsMan(models.Provider):
    """Meta-plugin for managing plugins"""

    name = "plugins"
    cli_name = "plugins"

    @cli.click.argument("plugin_name")
    @cli.click.option("--json-schema", "-s", is_flag=True, default=False)
    @cli.click.option("--markdown", "-m", is_flag=True, default=False)
    def schema(
        self, plugin_name, markdown: bool = False, json_schema: bool = True
    ) -> str:
        """Retrieves the configuration schema for the given plugin"""
        if not any([markdown, json_schema]):
            json_schema = True
        plugin = self.siblings[plugin_name]
        cfg = plugin.config
        model = cfg.__class__
        schema = cfg.schema()
        if json_schema:
            print(dumps.json(schema, indent=2))
        elif markdown:
            template_name = f"via {__file__}"
            t = render.get_template(
                from_string="""{% include 'pynchon/pydantic-model.md.j2' %}""",
                template_name=template_name,
            )
            fields = model.__fields__
            for dprop_name in model.get_properties():
                dprop = getattr(model, dprop_name, None)
                # import IPython; IPython.embed()
                if dprop:
                    fields[dprop_name] = dict(
                        is_dynamic=True,
                        annotation=str(dprop.fget.__annotations__.get("return", "")),
                        default=None,
                        required=False,
                        description=dprop.__doc__ or "(Missing docstring for property)",
                    )

            print(
                t.render(
                    plugin_name=plugin_name, model=model, fields=fields, schema=schema
                )
            )
        else:
            raise SystemExit(1)

    @cli.click.option("--name")
    @cli.click.option("--template-skeleton", "-t", is_flag=True, default=False)
    def new(self, name: str = None, template_skeleton: bool = False) -> None:
        """
        Create new plugin from template (for devs)
        """
        # FIXME: use pattern?
        plugins_d = abcs.Path(__file__).parents[0]
        template_plugin_f = plugins_d / "__template__.py"
        new_plugin_file = plugins_d / f"{name}.py"
        cmd = f"ls {new_plugin_file} || cp {template_plugin_f} {new_plugin_file} && git status"
        result = invoke(cmd, system=True)
        if template_skeleton:
            raise NotImplementedError()
        return result.succeeded

    @tagging.tags(click_aliases=["ls"])
    def list(self, **kwargs):
        """List all plugins"""
        return list(self.status()["plugins"].keys())

    @tagging.tags(click_aliases=["st", "stat"])
    def status(self) -> typing.Dict:
        """Returns details about all known plugins"""
        result = typing.OrderedDict()
        for name, p in self.siblings.items():
            result[name] = dict(priority=p.priority, key=p.get_config_key())
        return dict(plugins=result)



""" pynchon.plugins.project
"""

from fleks.util import tagging  # noqa

from pynchon import abcs, config, constants, models

from pynchon.cli import common, options  # noqa
from pynchon.util import files, lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class ProjectConfig(abcs.Config):
    """ """

    config_key: typing.ClassVar[str] = "project"
    shell_aliases: typing.Dict[str, str] = typing.Field(default={})
    subproject_patterns: typing.List[str] = typing.Field(default=[])

    exclude_patterns: typing.List[str] = typing.Field()

    # @tagging.tagged_property(conflict_strategy="override")
    @property
    def exclude_patterns(self):
        from pynchon.config import globals

        # globals = plugin_util.get_plugin("globals").get_current_config()
        global_ex = globals.exclude_patterns
        my_ex = self.__dict__.get("exclude_patterns", [])
        return list(set(global_ex + my_ex + ["**/pynchon/templates/includes/**"]))

    @property
    def name(self) -> typing.StringMaybe:
        """ """
        repo_name = config.git.repo_name
        return repo_name or abcs.Path(".").name

    @property
    def _workdir(self):
        return abcs.Path(".").absolute()

    @property
    def root(self) -> str:
        """ """
        git = config.GIT
        return constants.PYNCHON_ROOT or (git and git.root) or self._workdir

    @property
    def subproject(self) -> typing.Dict:
        """ """
        if constants.PYNCHON_ROOT:
            return {}
        git = config.GIT
        git_root = git["root"]
        r1 = self._workdir
        r2 = git_root and git_root.absolute()
        if r2 and (r1 != r2):
            rel_name = r1.relative_to(r2)
            LOGGER.debug(f"subproject detected: {rel_name}")
            # LOGGER.warning(f"subproject detected:\n\t({r1} != git[root] @ {r2})")
            return dict(name=self._workdir.name, root=self._workdir)
        return {}


class Project(models.Manager):
    """Meta-plugin for managing this project"""

    name = "project"
    priority = 2
    config_class = ProjectConfig

    @tagging.tags(click_aliases=["ls"])
    def list(self):
        """List subprojects associated with this project"""
        default = self[:"project"]
        proj_conf = self[:"project.subproject":default]
        project_root = proj_conf.get("root", None) or self[:"git.root":"."]
        # project_root = proj_conf.get("root", None) or '.'
        globs = [
            abcs.Path(project_root).joinpath("**/.pynchon.json5"),
        ]
        self.logger.warning(f"search patterns are {globs}")
        result = files.find_globs(globs)
        self.logger.debug(f"found {len(result)} subprojects (pre-filter)")
        excludes = self["exclude_patterns"]
        self.logger.debug(f"filtering search with {len(excludes)} excludes")
        result = [p for p in result if not p.match_any_glob(excludes)]
        result = [
            dict(
                name=str(p.relative_to(abcs.Path(".").absolute()).parent), config_file=p
            )
            for p in result
        ]
        result = [r for r in result if r["name"] != "."]
        self.logger.debug(f"found {len(result)} subprojects (post-filter)")
        return result

    # @common.kommand(
    #     name="version",
    #     parent=parent,
    #     formatters=dict(markdown=constants.T_VERSION_METADATA),
    #     options=[
    #         # FIXME: options.output_with_default('docs/VERSION.md'),
    #         options.format_markdown,
    #         options.output,
    #         options.header,
    #     ],
    # )
    # def project_version(format, output, header) -> None:
    #     """
    #     Describes version details for this package (and pynchon itself).
    #     """
    #     # from pynchon.api import python #, git
    #     import pynchon
    #     from pynchon.config import git, python
    #
    #     return dict(
    #         pynchon_version=pynchon.__version__,
    #         package_version=python.package.version,
    #         git_hash=git.hash,
    #     )



"""
"""

from pynchon import abcs, models  # noqa


class PythonPlanner(models.Planner):
    cli_label = "Python Tools"
    cli_description = (
        "Code transforms, docs, & boilerplate generation for Python projects"
    )



""" pynchon.plugins.mkdocs
"""

import urllib
from pathlib import Path

import yaml
import fleks
from fleks import tagging

from pynchon.plugins import util as plugin_util
from pynchon.util.text import loadf

from pynchon import abcs, api, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class MkdocsPluginConfig(abcs.Config):
    config_key: typing.ClassVar[str] = "mkdocs"
    config_file: str = typing.Field(default=None)

    @property
    def tags(self) -> typing.List:
        """ """
        tags = set()
        for p in self.pages:
            tags = tags.union(set(p.get("tags", [])))
        # NB: removes empty-string
        return sorted(list(filter(None, tags)))

    @property
    def drafts(self):
        ignore_list = ["index.md", "tags.md", "nav.md"]
        return [
            p
            for p in self.pages
            if all([p["draft"], p["path"].name not in ignore_list])
        ]

    @property
    def pages(self) -> typing.List:
        """ """
        from mkdocs.structure.files import File
        from mkdocs.structure.pages import Page

        pages = []
        mconf = self.config
        if mconf:
            ddir = abcs.Path(mconf.get("docs_dir", "docs"))
            from mkdocs.config.defaults import MkDocsConfig

            cfg = MkDocsConfig()
            data = yaml.load(open(self.config_file).read(), yaml.FullLoader)
            cfg.load_dict(data)
            cfg.validate()  # fl = File(
            #     'heredoc/ambient-calculus-1.md',
            #     cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls)
            # config_file
            # pconf = mconf['plugins'] if 'plugins' in mconf else {}
            # bconf = [conf for conf in pconf if 'blogging' in list(conf.keys())]
            # bconf = bconf[0]['blogging'] if bconf else {}
            pfiles = ddir.glob("**/*.md")
            # pfiles = [p.relative_to(ddir) for p in pfiles]
            for pfile in pfiles:
                rel_pfile = pfile.relative_to(ddir)
                mfile = File(
                    str(rel_pfile), cfg.docs_dir, cfg.site_dir, cfg.use_directory_urls
                )
                pg = Page(
                    file=mfile,
                    config=cfg,
                    title=None,
                )
                pg.read_source(cfg)
                tags = pg.meta.get("tags", [])
                draft = any([pg.meta.get("draft", False), "draft" in str(rel_pfile)])
                #   {%- for p in mkdocs.blog_posts|default([]) %}{% set p=p|Path %}{% set p=p.relative_to(mkdocs.config.docs_dir | default('docs/')) %}
                # * [{{p.stem.replace('-',' ').title()}}]({{p}}){%- endfor -%}
                pmeta = dict(
                    title=pg.title,
                    # relative_url=pg.url,
                    relative_url=f"{self.site_relative_url}/{pg.url}",
                    path=pfile.absolute(),
                    rel_path=str(rel_pfile),
                    tags=tags,
                    # thing=rel_pfile.stem.replace('- ',' ').title(),
                    draft=draft,
                )
                pages.append(pmeta)
        return pages

    @property
    def blog_posts(self) -> list:
        """
        returns blog posts, iff blogging plugin is installed.
        resulting files, if any, will not include index and
        will be sorted by modification time
        """
        ignore_list = ["index.md", "tags.md", "nav.md"]
        return [
            p
            for p in self.pages
            if all([not p["draft"], p["path"].name not in ignore_list])
        ]
        # mconf = self.config
        # if mconf:
        #     pconf = mconf["plugins"] if "plugins" in mconf else {}
        #     ddir = abcs.Path(mconf.get("docs_dir", "docs"))
        #     bconf = [conf for conf in pconf if "blogging" in list(conf.keys())]
        #     bconf = bconf[0]["blogging"] if bconf else {}
        #     blog_dirs = [ddir / bdir for bdir in bconf.get("dirs", [])]
        #     result = []
        #     for bdir in blog_dirs:
        #         result += [g for g in bdir.glob("**/*.md") if g.name not in ignore_list]
        #     result = reversed(sorted(result, key=lambda p: p.lstat().st_mtime))
        #     result = [str(p) for p in result]
        #     return result

    @property
    def site_relative_url(self):

        site_url = self.config["site_url"] if "site_url" in self.config else None
        if site_url:
            return urllib.parse.urlparse(site_url).path

    @property
    def config(self) -> typing.Dict:
        """
        returns a dictionary with the current mkdocs configuration
        """
        fname = self.config_file
        if fname is None:
            return {}

        return loadf.yaml(fname)

    @property
    def config_file(self) -> typing.StringMaybe:
        """returns the path to the mkdocs config-file, if applicable"""
        docs = plugin_util.get_plugin("docs", strict=False)
        docs = docs and docs.get_current_config()
        subproject = plugin_util.get_plugin("subproject", strict=False)
        subproject = subproject and subproject.get_current_config()
        project = plugin_util.get_plugin("project", strict=False)
        project = project and project.get_current_config()
        candidates = filter(
            None,
            [
                abcs.Path(".").absolute(),
                docs and docs.root,
                subproject and subproject.root,
                project and project.root,
            ],
        )
        for folder in [Path(c) for c in candidates]:
            cand = folder / "mkdocs.yml"
            if cand.exists():
                return str(cand.absolute())


DEFAULT_LOG_FILE = ".tmp.mkdocs.log"


class Mkdocs(models.Planner):
    """Mkdocs helper"""

    priority = 8  # before mermaid
    name = "mkdocs"
    cli_name = "mkdocs"
    cli_label = "Docs Tools"
    config_class = MkdocsPluginConfig

    @property
    def config_file(self) -> str:
        return self["config_file"]

    def serve(self, background: bool = True):
        """
        Wrapper for `mkdocs serve`
        """
        from pynchon.util.os import invoke

        bg = "&" if background else ""
        cmd = f"mkdocs serve --config-file {self.config_file} >> {DEFAULT_LOG_FILE} 2>&1 {bg}"
        result = invoke(cmd)
        return result

    @tagging.tags(click_aliases=["ls"])
    def list(self):
        """Lists site-pages based on mkdocs.yml"""
        return self.config.pages

    @fleks.cli.click.argument("files", nargs=-1)
    def open(
        self,
        files: tuple = tuple(),
    ):
        """
        Opens `dev_addr` in a webbrowser
        """
        import webbrowser

        file = files[0] if files else "."
        # index_f = Path(self.site_dir).absolute() / "index.html"
        # url = f"file://{index_f}"
        mconfig = self.config.config
        url = mconfig["dev_addr"]
        url = f"{url}/{file}" if file else url
        url = f"http://{url}" if not url.startswith("http") else url
        self.logger.warning(f"opening {url}")
        return webbrowser.open(url)

    @property
    def site_dir(self) -> str:
        """
        Returns mkdocs `site_dir` if present in config, or guesses what it should be
        """
        plugin_cfg = self.config
        mkdocs_config = plugin_cfg.config
        result = str(mkdocs_config.get("site_dir", self.working_dir / "site"))
        self.logger.warning(f"returning {result}")
        return result

    # def _hook_open_after_apply(self, result) -> bool:
    #     raise Exception(result)

    def plan(self):
        """
        Runs a plan for this plugin
        """
        plan = super(self.__class__, self).plan()
        plan.append(
            self.goal(
                type="render",
                resource=self.site_dir,
                command=f"mkdocs build --config-file {self.config_file}",
            )
        )
        return plan



""" pynchon.models.planner """

import typing

from fleks import tagging
from memoized_property import memoized_property
from fleks.util.tagging import tags

from pynchon import abcs, cli
from pynchon.app import app

from . import planning
from .plugins import BasePlugin

from pynchon.util import files, lme, typing  # noqa


LOGGER = lme.get_logger(" ")


@tags(cli_label="Planner")
class AbstractPlanner(BasePlugin):
    """A plugin-type that provides plan/apply basics"""

    cli_label = "Planner"

    def _list(self, use_glob=None, changes=False):
        """
        Lists affected resources for this project
        """
        use_glob = use_glob or getattr(self.config, "file_glob", None)
        assert use_glob
        default = self[:"project"]
        proj_conf = self[:"project.subproject":default]
        project_root = proj_conf.get("root", None) or self[:"git.root":"."]
        globs = [
            abcs.Path(project_root).joinpath(f"**/{use_glob}"),
        ]
        self.logger.debug(f"search patterns are {globs}")
        result = files.find_globs(globs)
        self.logger.debug(f"found {len(result)} {use_glob} files (pre-filter)")
        excludes = self["exclude_patterns"::[]]
        # NB: always honor the global-excludes
        excludes += self.siblings["globals"]["exclude_patterns"::[]]
        self.logger.debug(f"filtering search with {len(excludes)} excludes")
        result = [p for p in result if not p.match_any_glob(excludes)]
        self.logger.debug(f"found {len(result)} {use_glob} files (post-filter)")
        if not result:
            err = f"active, but found no {use_glob} files!"
            self.logger.critical(err)
        return result

    @tags(publish_to_cli=False)
    def goal(self, **kwargs):
        """ """
        return planning.Goal(
            owner=f"{self.__class__.__module__}.{self.__class__.__name__}", **kwargs
        )

    @property
    def Plan(self):
        return planning.Plan

    def plan(
        self,
        config=None,
        goals=[],
    ) -> planning.Plan:
        """
        Creates a plan for this plugin
        """
        # app.manager.status_bar.update(app='PLAN')
        app.status_bar.update(app="Pynchon::PLAN", stage=f"plugin:{self.name}")
        plan = self.Plan(owner=self.name)
        if not goals:
            goals = getattr(self.config, "goals", [])
            goals and LOGGER.critical(f"{self.name} goals from config: {goals}")
            for g in goals:
                _type = g.pop("type", "from-config")
                cmd = g.pop("command", None)
                cmd = cmd and cmd.format(**g)
                g.update(command=cmd, type=_type)
                plan.append(self.goal(**g))
        else:
            self.logger.critical(f"packing  {len(goals)} goals")
            for g in goals:
                plan.append(g)
        app.status_bar.update(app="Pynchon", stage=f"{len(plan)}")
        return plan

    def _dispatch_apply_hooks(self, results: planning.ApplyResults):
        """ """
        # write status event (used by the app-console)
        app.status_bar.update(
            app="Pynchon::HOOKS",
            stage=f"{self.__class__.name}",
        )
        if results.finished:
            hooks = self.apply_hooks
            if hooks:
                LOGGER.warning(
                    f"{self.name}.apply: dispatching {len(hooks)} hooks: {hooks}"
                )
                hook_results = []
                for hook in hooks:
                    hook_results.append(self.run_hook(hook, results))
            else:
                LOGGER.warning(f"{self.name}.apply: no hooks to run.")
        else:
            LOGGER.critical(f"{self.name}: Skipping hooks: ")
            LOGGER.critical(
                f"  {len(results.goals)-len(results.actions)} goals incomplete"
            )

    @cli.options.quiet
    @cli.options.parallelism
    @cli.options.fail_fast
    def apply(
        self,
        plan: planning.Plan = None,
        parallelism: str = "0",
        fail_fast: bool = False,
        quiet: bool = False,
    ) -> planning.ApplyResults:
        """
        Executes the plan for this plugin
        """
        parallelism = int(parallelism)
        # app.status_bar.update(
        #     app="Pynchon::APPLY",
        #     stage=f"plugin:{self.__class__.name}"
        # )
        plan = plan or self.plan()
        LOGGER.critical(
            f"{self.name}.apply ( {len(plan)} goals with {parallelism} workers)"
        )
        results = plan.apply(parallelism=parallelism, git=self.siblings["git"])

        LOGGER.critical(
            f"{self.name}.apply finished ( {len(results.actions)}/{len(results.goals)} goals )"
        )
        self._dispatch_apply_hooks(results)
        if not quiet:
            return results

    def _validate_hooks(self, hooks):
        # FIXME: validation elsewhere
        for x in hooks:
            assert isinstance(x, (str,))
            assert " " not in x
            assert "_" not in x
            assert x.strip()

    @memoized_property
    def apply_hooks(self):
        """ """
        hooks = [x for x in self.hooks if x.split("-")[-1] == "apply"]
        apply_hooks = self["apply_hooks"::[]]
        hooks += [
            x + ("-apply" if not x.endswith("-apply") else "") for x in apply_hooks
        ]
        hooks = list(set(hooks))
        self._validate_hooks(hooks)
        return hooks

    @memoized_property
    def hooks(self):
        """ """
        hooks = self["hooks"::[]]
        self._validate_hooks(hooks)
        return hooks

    def _hook_diff_after_apply(self, result: planning.ApplyResults) -> bool:
        """ """
        self.logger.warning("diff-after-apply not implemented yet")

    def _hook_open_after_apply(self, result: planning.ApplyResults) -> bool:
        """ """
        changes = list({r.resource for r in result})
        changes = [abcs.Path(rsrc) for rsrc in changes]
        changes = [rsrc for rsrc in changes if not rsrc.is_dir()]
        self.logger.warning(f"Opening {len(changes)} changed resources.")
        docs_plugin = self if self.name == "docs" else self.siblings["docs"]
        for ch in changes:
            docs_plugin.open(ch)
        return True

    @typing.validate_arguments
    def run_hook(self, hook_name: str, results: planning.ApplyResults):
        """
        :param hook_name: str:
        :param results: planning.ApplyResults:
        """

        class HookNotFound(Exception):
            pass

        class HookFailed(RuntimeError):
            pass

        norml_hook_name = hook_name.replace("-", "_")
        fxn_name = f"_hook_{norml_hook_name}"
        hook_fxn = getattr(self, fxn_name, None)
        if hook_fxn is None:
            err = [self.__class__, [hook_name, fxn_name]]
            self.logger.critical(err)
            raise HookNotFound(err)
        hook_result = hook_fxn(results)
        self.logger.critical(hook_result)
        return hook_result


class ShyPlanner(AbstractPlanner):
    """ShyPlanner uses plan/apply workflows, but they must be
    executed directly.  ProjectPlugin (or any other parent plugins)
    won't include this as a sub-plan.
    """

    contribute_plan_apply = False


@tags(cli_label="Manager")
class Manager(ShyPlanner):
    cli_label = "Project Tools"
    cli_description = "Tools for project management"


class ResourceManager(Manager):
    @property
    def changes(self):
        """Set(git_changes).intersection(plugin_resources)"""
        git = self.siblings["git"]
        changes = git.modified
        these_changes = set(changes).intersection(set(self.list(changes=False)))
        return dict(modified=list(these_changes))

    @tagging.tags(click_aliases=["ls"])
    @cli.click.option(
        "--changes",
        "-m",
        "changes",
        is_flag=True,
        default=False,
        help="returns the git-modified subset",
    )
    def list(self, changes: bool = False):
        """Lists resources associated with this plugin"""
        if changes:
            return self.changes["modified"]
        from pynchon import abcs
        from pynchon.util import files

        try:
            include_patterns = self["include_patterns"]
            root = self["root"]
        except (KeyError,) as exc:
            self.logger.critical(
                f"{self.__class__} tried to use self.list(), but does not follow protocol"
            )
            self.logger.critical(
                "self['include_patterns'] and self['root'] must both be defined!"
            )
            raise
        root = abcs.Path(root)
        # proot = self.project_config['pynchon']['root']
        tmp = [p for p in include_patterns if abcs.Path(p).is_absolute()]
        tmp += [root / p for p in include_patterns if not abcs.Path(p).is_absolute()]
        # tmp += [proot / p for p in include_patterns if not abcs.Path(p).is_absolute()]
        return files.find_globs(tmp)


class Planner(ShyPlanner):
    """Planner uses plan/apply workflows, and contributes it's plans
    to ProjectPlugin (or any other parent plugins).
    """

    contribute_plan_apply = True
    cli_description = ""


class RenderingPlugin(Planner):
    cli_label = "Rendering Tools"
    cli_description = ""



""" pynchon.plugins.pandoc
"""

from fleks import cli

from fleks.util import tagging  # noqa

from pynchon import abcs, events, models  # noqa
from pynchon.util import lme, typing  # noqa

LOGGER = lme.get_logger(__name__)


class Pandoc(models.DockerWrapper, models.Planner):
    """
    Wrapper around `pandoc` docker image
    """

    class config_class(models.DockerWrapper.BaseConfig):
        config_key: typing.ClassVar[str] = "pandoc"
        docker_args: typing.List = typing.Field(
            default=["--toc", "--variable fontsize=10pt"]
        )
        docker_image: str = typing.Field(default="pandoc/extra:latest")
        goals: typing.List[typing.Dict] = typing.Field(default=[], help="")

    name = "pandoc"
    cli_name = "pandoc"
    cli_label = "Docs Tools"
    # contribute_plan_apply = False

    @cli.click.command(
        context_settings=dict(
            ignore_unknown_options=True,
            allow_extra_args=True,
        )
    )
    def pdflatex(self, *args, **kwargs):
        # command = self._get_docker_command(self.command_extra)
        command = self._get_docker_command_base(
            self.command_extra,
            # docker_args='-it',
            entrypoint="pdflatex",
        )
        LOGGER.warning(command)
        result = self._run_docker(command)
        # result = self.apply(plan=super().plan(
        #     goals=[
        #         self.goal(
        #             type="?",
        #             # resource=kwargs.get("output", kwargs.get("o", None)),
        #             command=command,
        #         )
        #     ]
        # ))
        # if result.ok:
        #     raise SystemExit(0)
        # else:
        #     LOGGER.critical(f"Action failed: {result.actions[0].dict()}")
        #     raise SystemExit(1)

    def shell(self):
        """Starts interactive shell for pandoc container"""
        command = ""
        command = self._get_docker_command_base(docker_args="-it", entrypoint="sh")
        LOGGER.warning(command)
        result = self.apply(
            plan=super().plan(
                goals=[
                    self.goal(
                        type="interactive",
                        # resource=kwargs.get("output", kwargs.get("o", None)),
                        command=command,
                    )
                ]
            )
        )
        if result.ok:
            raise SystemExit(0)
        else:
            LOGGER.critical(f"Action failed: {result.actions[0].error}")
            raise SystemExit(1)

    @tagging.tags(click_aliases=["markdown.to-pdf"])
    @cli.options.output_file
    @cli.click.argument("file")
    def md_to_pdf(
        self,
        file: str = None,
        output: str = None,
    ):
        """
        Converts markdown files to PDF with pandoc
        """
        output = abcs.Path(output or f"{abcs.Path(file).stem}.pdf")
        # cmd = f"docker run -v `pwd`:/workspace -w /workspace {docker_image} {file} {docker_args} -o {output}"
        return self.run(file, output=output)
        # plan = super().plan(
        #     goals=[
        #         self.goal(resource=output.absolute(),
        #         type="render", command=cmd)]
        # )
        # return self.apply(plan=plan)



""" pynchon.plugins.Core
"""

import fleks
from fleks import tagging

from pynchon import abcs, api, cli, models
from pynchon.bin import entry
from pynchon.core import Config as CoreConfig
from pynchon.util import lme
from pynchon.models import planning

LOGGER = lme.get_logger(" ")

classproperty = fleks.util.typing.classproperty


@tagging.tags(click_aliases=["c"])
class Core(models.Planner):
    """Core Plugin"""

    name = "core"
    priority = -1
    config_class = CoreConfig

    # NB: prevents recursion when `pynchon plan` is used!
    contribute_plan_apply = False

    @classproperty
    def click_group(kls):
        """ """
        kls._finalized_click_groups[kls] = entry
        return kls._finalized_click_groups[kls]

    @classmethod
    def get_current_config(kls):
        """ """
        from pynchon import config as config_mod  # noqa

        result = getattr(config_mod, kls.get_config_key())
        return result

    def cfg(self):
        """Show current project config (with templating/interpolation)"""
        tmp = self.project_config
        return tmp

    def _init_config(self, config_path):
        """ """
        cfg = self.cfg().dict()
        pop_list = ["apply_hooks"]
        for plugin_name in cfg:
            try:
                sib = self.siblings[plugin_name]
            except (KeyError,):
                sib = None
            else:
                if isinstance(sib, (models.Provider,)):
                    pop_list.append(plugin_name)
        for p in pop_list:
            cfg.pop(p)
        LOGGER.critical(f"Writing {config_path.absolute()}")
        from pynchon.util.text import dumpf

        dumpf.json(cfg, file=config_path)

    def init(self, strict=True):
        """Write .pynchon.json5 if it's not present"""
        default_config = abcs.Path(".") / ".pynchon.json5"
        if default_config.exists():
            LOGGER.critical(f"{default_config} exists, nothing to do.")
            err = f"Refusing to create {default_config} (file already exists)"
            LOGGER.critical(err)
            if strict:
                raise SystemExit(1)
        else:
            self._init_config(default_config)
            return True

    # @cli.click.flag("--bash", help="bootstrap bash")
    @cli.click.flag("--bashrc", help="bootstrap bashrc")
    # @cli.click.flag("--bash-completions", help="bootstrap completions")
    @cli.click.flag("--pynchon", help="bootstrap .pynchon.json5")
    # @cli.click.flag("--makefile", help="bootstrap Makefile")
    # @cli.click.flag("--tox", help="bootstrap tox")
    def bootstrap(
        self,
        pynchon: bool = False,
        bashrc: bool = False,  # noqa
        bash: bool = False,
        bash_completions: bool = False,
        makefile: bool = False,
        tox: bool = False,
    ) -> None:
        """
        Helpers for bootstraping various project boilerplate.
        """
        template_prefix = f"{self.plugin_templates_prefix}/bootstrap"
        pynchon_completions_script = ".tmp.pynchon.completions.sh"
        bashrc_snippet = ".tmp.pynchon.bashrc"
        if bash_completions:
            pass

            gr = self.__class__.click_entry
            all_known_subcommands = [
                " ".join(x.split()[1:])
                for x in cli.click.subcommand_tree(
                    gr, mode="text", path=tuple(["pynchon"]), hidden=False
                ).keys()
            ]
            # head = [x for x in all_known_subcommands if len(x.split()) == 1]
            # rest = [x for x in all_known_subcommands if x not in head]
            # tmp = collections.defaultdict(list)
            # for phrase in rest:
            #     bits = phrase.split()
            #     k = bits.pop(0)
            #     tmp[k] += bits
            # print(list(tmp.keys()))
            print(all_known_subcommands)
            return
        #   if bash:
        #       rest = [
        #           f"""    '{k}'*)
        #         while read -r; do COMPREPLY+=( "$REPLY" ); done < <( compgen -W "$(_pynchon_completions_filter "{' '.join(subs)}")" -- "$cur" )
        #         ;;
        #       """
        #           for k, subs in tmp.items()
        #       ]
        #       rest += [
        #           f"""    *)
        # while read -r; do COMPREPLY+=( "$REPLY" ); done < <( compgen -W "$(_pynchon_completions_filter "{' '.join(head)}")" -- "$cur" )
        # ;;"""
        #       ]
        #       # LOGGER.warning("This is intended to be run through a pipe, as in:")
        #       # LOGGER.critical("pynchon bootstrap --bash | bash")
        #       tmpl = api.render.get_template(f"{template_prefix}/bash.sh")
        #       content = tmpl.render(head=head, rest="\n".join(rest), **self.config)
        #       files.dumps(content=content, file=pynchon_completions_script)
        #       LOGGER.warning(
        #           f"To refresh your shell, run: `source {pynchon_completions_script}`"
        #       )
        #       return dict()
        # if bashrc:
        #     LOGGER.critical(
        #         "To use completion hints every time they are "
        #         "present in a folder, adding this to .bashrc:"
        #     )
        #     tmpl = api.render.get_template(f"{template_prefix}/bashrc.sh")
        #     content = tmpl.render(pynchon_completions_script=pynchon_completions_script)
        #     files.dumps(content=content, file=bashrc_snippet, logger=LOGGER.info)
        #     return files.block_in_file(
        #         target_file=abcs.Path("~/.bashrc").expanduser(),
        #         block_file=bashrc_snippet,
        #     )
        elif pynchon:
            return self.init()
        elif bash:
            this_cmd = "pynchon bootstrap --bash"  # FIXME: get from click-ctx
            LOGGER.debug("collecting `shell_aliases` from all plugins")
            out = self.siblings.collect_config_dict("shell_aliases")
            out = "\n".join([f"alias {k}='{v}';" for k, v in out.items()])
            print(out)
            LOGGER.warning(f'for this session, use "source <({this_cmd})"')
            LOGGER.warning(f'for it to be permanent, use "{this_cmd} >> ~/.bashrc"')
        elif tox or makefile:
            tail = "Makefile" if makefile else "tox.ini"
            tmpl = api.render.get_template(f"{template_prefix}/{tail}")
            content = tmpl.render(**self.project_config.dict())
            print(content)

    def raw(self) -> None:
        """
        Returns (almost) raw config, without interpolation
        """
        from pynchon.config import RAW  # noqa

        print(RAW.json())

    @property
    def sorted_plugins(self):
        plugins = self.siblings.values()
        plugins = [
            p
            for p in plugins
            if isinstance(p, models.AbstractPlanner) and p.contribute_plan_apply
        ]
        plugins = sorted(plugins, key=lambda p: p.priority)
        return plugins

    @cli.options.quiet
    def plan(
        self,
        config=None,
        quiet: bool = False,
        flattened: bool = True,
    ) -> models.Plan:
        """Runs planning for all plugins"""

        config = config or self.project_config
        plan = super(self.__class__, self).plan(config)
        plans = [] if not flattened else None
        plugins = self.sorted_plugins
        self.logger.critical(f"Planning on behalf of: {[p.name for p in plugins]}")
        for plugin_obj in plugins:
            subplan = plugin_obj.plan()
            if not subplan:
                self.logger.warning(
                    f"subplan for '{plugin_obj.__class__.__name__}' is empty!"
                )
            else:
                if flattened:
                    for g in subplan.goals:
                        self.logger.info(f"{plugin_obj} contributes {g}")
                        plan.append(g)
                else:
                    plans.append(subplan)
        result = plan.finalize() if flattened else plans
        return result if not quiet else None

    @cli.click.option("--parallelism", "-p", default="1", help="Paralellism")
    @cli.click.flag("--fail-fast", "-f", default=False, help="fail fast")
    @cli.options.quiet
    def apply(
        self,
        plan: planning.Plan = None,
        parallelism: str = "1",
        quiet: bool = False,
        fail_fast: bool = False,
    ) -> planning.ApplyResults:
        """
        Executes the plan for all plugins
        """
        parallelism = int(parallelism)
        plans = plan or self.plan(flattened=False)
        for plan in plans:
            results = plan.apply(parallelism=parallelism, git=self.siblings["git"])
            LOGGER.critical(
                f"Finished apply for subplan ({len(results.actions)}/{len(results.goals)} goals)"
            )
        # self.dispatch_apply_hooks(results)
        return results if not quiet else None



""" pynchon.cli.options
"""

from pynchon.cli import click

quiet = click.flag("--quiet", "-q", default=False, help="Disable JSON output")
fail_fast = click.flag("--fail-fast", "-f", default=False, help="fail fast")
parallelism = click.option(
    "--parallelism",
    "-p",
    default="0",
    help="Set number of workers (or 0 for sequential action)",
)



""" pynchon.models.plugins.docker
"""

from pynchon import abcs
from pynchon.cli import click

from .tool import ToolPlugin

from pynchon.util import lme, typing  # noqa

from . import validators  # noqa

LOGGER = lme.get_logger("DOCKER")


class DockerWrapper(ToolPlugin):
    """
    Wrappers for dockerized tools
    """

    class BaseConfig(abcs.Config):
        docker_image: str = typing.Field(default="docker/hello-world")
        docker_args: typing.List = typing.Field(default=[])

    cli_label = "Docker Wrappers"
    cli_description = "Plugins that wrap invocations on containers"
    contribute_plan_apply = False
    priority = 2
    __class_validators__ = [
        validators.require_conf_key,
    ]

    @property
    def command_extra(self):
        import sys

        command = sys.argv[sys.argv.index(self.click_group.name) + 2 :]
        return " ".join(command)

    def _get_docker_command_base(self, *args, **kwargs):
        docker_image = kwargs.pop("docker_image", self["docker_image"])
        docker_args = kwargs.pop("docker_args", "")
        docker_args = (
            docker_args + " " + " ".join(f'--{k}="{v}"' for k, v in kwargs.items())
        )
        return (
            "docker run -v `pwd`:/workspace -w /workspace "
            f"{docker_args} {docker_image} {' '.join(args)}"
        )

    def _get_docker_command(self, *args, **kwargs):
        """ """
        cmd_t = self._get_docker_command_base(" ".join(args))
        docker_args = " ".join(self["docker_args"])
        zip_kws = " ".join(["{k}={v}" for k, v in kwargs.items()])
        cmd_t += f" {docker_args} {zip_kws}"
        return cmd_t

    @click.command(
        context_settings=dict(
            ignore_unknown_options=True,
            allow_extra_args=True,
        )
    )
    def run(self, *args, **kwargs):
        """Passes given command through to docker-image this plugin wraps"""
        command = self._get_docker_command(self.command_extra)
        LOGGER.warning(command)
        plan = super().plan(
            goals=[
                self.goal(
                    type="render",
                    resource=kwargs.get("output", kwargs.get("o", "")),
                    command=command,
                )
            ]
        )
        result = self.apply(plan=plan)
        if result.ok:
            raise SystemExit(0)
        else:
            LOGGER.critical(f"Action failed: {result.actions[0].error}")
            raise SystemExit(1)

    def _run_docker(self, cmd, **kwargs):
        """ """
        from pynchon.util.os import invoke

        LOGGER.critical(cmd)
        result = invoke(cmd, **kwargs)
        LOGGER.warning(result.stdout)
        return result


class DiagramTool(DockerWrapper):
    cli_label = "Diagramming Tools"
    cli_description = "View and render technical diagrams from source, in several formats.  (Usually these require docker, but no other system dependencies.)"



""" pynchon.plugins.markdown
"""

import marko
from fleks import tagging

from pynchon import abcs, api, cli, events, models  # noqa
from pynchon.util import files, lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)

ElementList = typing.List[typing.Dict]


class Markdown(models.DockerWrapper, models.Planner):
    """Markdown Tools"""

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "markdown"
        goals: typing.List[str] = typing.Field(
            default=[], description="Extra goals related to markdown"
        )
        include_patterns: typing.List[str] = typing.Field(
            default=[], description="Patterns to include"
        )
        exclude_patterns: typing.List[str] = typing.Field(
            default=[], description="File globs to exclude from listing"
        )
        root: typing.Union[str, abcs.Path, None] = typing.Field(
            default=None, description=""
        )
        linter_docker_image: str = typing.Field(
            default="peterdavehello/markdownlint",
            description="Container to use for markdown linter",
        )
        linter_args: typing.List[str] = typing.Field(
            description="Arguments to pass to `linter_docker_image`",
            default=[
                "--disable MD013",  # line-length
                "--disable MD045",  # Images should have alternate text
                "--disable MD033",  # Allow HTML
                "--disable MD041",  # first-line-h1
                "--disable MD042",  # No empty links
                "--fix",
            ],
        )
        goals: typing.List[typing.Dict] = typing.Field(default=[], description="")

    name = "markdown"
    cli_name = "markdown"
    cli_label = "Docs Tools"
    priority = 0

    @tagging.tags(click_aliases=["ls"])
    def list(self, changes=False):
        """
        Lists affected resources (**.md) for this project
        """
        default = self[:"project"]
        proj_conf = self[:"project.subproject":default]
        project_root = proj_conf.get("root", None) or self[:"git.root":"."]
        # project_root = proj_conf.get("root", None) or '.'
        globs = [
            abcs.Path(project_root).joinpath("**/*.md"),
        ]
        self.logger.debug(f"search patterns are {globs}")
        result = files.find_globs(globs)
        self.logger.debug(f"found {len(result)} j2 files (pre-filter)")
        excludes = self["exclude_patterns"]
        self.logger.debug(f"filtering search with {len(excludes)} excludes")
        result = [p for p in result if not p.match_any_glob(excludes)]
        self.logger.debug(f"found {len(result)} j2 files (post-filter)")
        if not result:
            err = f"{self.__class__.__name__} is active, but found no .j2 files!"
            self.logger.critical(err)
        return result

    @cli.click.argument("paths", nargs=-1)
    def normalize(self, paths):
        """Use `markdownlint` to normalize input paths"""
        docker_image = self["linter_docker_image"]
        linter_args = " ".join(self["linter_args"])
        goals = []
        for path in paths:
            goals.append(
                self.goal(
                    resource=path,
                    type="normalize",
                    command=(
                        f"docker run -v `pwd`:/workspace "
                        f"-w /workspace {docker_image} "
                        f"markdownlint {linter_args} {path}"
                    ),
                )
            )
        return self.apply(plan=self.plan(goals=goals))

    @cli.click.argument("paths", nargs=-1)
    @tagging.tags(click_aliases=["show"])
    def preview(self, paths):
        """
        Previews markdown in the terminal
        """
        # FIXME?: rich doesn't link hypertext.  patch upstream?
        from rich.console import Console
        from rich.markdown import Markdown

        console = Console()
        for p in paths:
            with open(p) as fhandle:
                md = Markdown(fhandle.read())
                console.print(md)
        # FIXME: glow is awesome but using it from docker seems to strip color
        # viewer_docker_image: str = typing.Field(
        #     default='charmcli/glow',
        #     help="Container to use for markdown console viewer")
        # docker_image = self["viewer_docker_image"]
        #         # viewer_args = " ".join(self["viewer_args"])
        #         return self._run_docker(
        #                         f"docker run -t -v `pwd`:/workspace "
        #                         f"-w /workspace {docker_image} "
        #                         f" {' '.join(paths)}"
        #                     )

    @cli.click.flag("-p", "--python", help="only python codeblocks")
    @cli.click.flag("-b", "--bash", help="only bash codeblocks")
    @cli.click.argument("file")
    def doctest(
        self,
        file: str = None,
        python: bool = False,
        bash: bool = False,
    ) -> ElementList:
        """Runs doctest for fenced code inside the given markdown files"""
        assert python or bash
        element_lst = self.parse(file=file, python=python, bash=bash)
        if not element_lst:
            LOGGER.critical(f"filtered element list is empty! {element_lst}")

        def _doctest(element):
            LOGGER.critical(element)
            child = element["children"][0]
            assert child["element"] == "raw_text"
            script: str = child["children"]
            raise Exception(script)
            # return #shil.invoke(script,...))

        for el in element_lst:
            el.update(_doctest(el))
        return element_lst

    @tagging.tags(click_aliases=["parse.markdown"])
    @cli.click.flag("-c", "--codeblocks", help="only codeblocks")
    @cli.click.flag("-p", "--python", help="only python codeblocks")
    @cli.click.flag("-b", "--bash", help="only bash codeblocks")
    @cli.click.flag("-l", "--links", help="only links")
    @cli.click.flag("--all", "-a", help="run for each file found by `list`")
    @cli.click.argument("files", nargs=-1)
    def parse(
        self,
        files: typing.Tuple = tuple(),
        all: bool = False,
        codeblocks: bool = False,
        python: bool = False,
        links: bool = False,
        bash: bool = False,
    ) -> ElementList:
        """Parses given markdown file into JSON"""
        from bs4 import BeautifulSoup  # noqa
        from marko.ast_renderer import ASTRenderer  # noqa

        codeblocks = codeblocks or python or bash
        assert files or all and not (files and all)
        if files:
            files = list(files)
        else:
            files = self.list()
            LOGGER.warning(f"parsing all markdown from: {files} ")
        out = {}
        for file in files:
            LOGGER.warning(f"parsing: {file}")
            file = str(file)
            with open(file) as fhandle:
                content = fhandle.read()
            if links:
                parsed = marko.Markdown()(content)
                soup = BeautifulSoup(parsed, features="html.parser")
                out[file] = []
                for a in soup.find_all("a", href=True):
                    this_link = a["href"]
                    if this_link.strip() == "#":
                        LOGGER.warning(f"{file}: has placeholder link '#' ")
                    else:
                        out[file] += [this_link]
            else:
                parsed = marko.Markdown(renderer=ASTRenderer)(content)
                children = parsed["children"]
                out[file] = []
                for child in children:
                    if child.get("element") == "fenced_code":
                        lang = child.get("lang")
                        if lang is not None:
                            out[file] += [child]
                LOGGER.critical(child)
                if python:
                    out[file] += [ch for ch in out if child.get("lang") == "python"]
                if bash:
                    out[file] += [ch for ch in out if child.get("lang") == "bash"]
        return {k: v for k, v in out.items() if v}



""" pynchon.plugins.drawio

    A Wrapper for docker-containers that
    provide the "drawio" diagramming utility

    Live Cloud Version:
        https://app.diagrams.net/
    Local Server:
        https://hub.docker.com/r/jgraph/drawio
        https://www.drawio.com/blog/diagrams-docker-app
    CLI Tools:
        https://github.com/rlespinasse/docker-drawio-desktop-headless
"""

import webbrowser

from fleks import tagging

from pynchon import abcs, api, cli, events, models  # noqa
from pynchon.util import files, lme, text, typing  # noqa

LOGGER = lme.get_logger(__name__)

ElementList = typing.List[typing.Dict]
DEFAULT_HTTP_PORT = 8080
DEFAULT_DOCKER_NAME = "drawio-server"


@tagging.tags(click_aliases=["draw"])
class DrawIO(models.DiagramTool, models.Planner):
    """
    Wrapper for docker-containers that
    provide the "drawio" diagramming utility
    """

    class config_class(abcs.Config):
        config_key: typing.ClassVar[str] = "drawio"
        file_glob: str = typing.Field(
            default="*.drawio", description="Where to find jinja templates"
        )

        docker_image: str = typing.Field(
            default="jgraph/drawio", help="Docker image to use"
        )
        http_port: str = typing.Field(help="Port to use", default=DEFAULT_HTTP_PORT)
        docker_args: typing.List = typing.Field(
            default=[f"-it --rm --name={DEFAULT_DOCKER_NAME}"],
            help="Docker args to use",
        )
        export_docker_image: str = typing.Field(
            default="rlespinasse/drawio-desktop-headless"
        )
        export_width: int = typing.Field(help="", default=800)
        export_args: typing.List = typing.Field(
            help="",
            default=[
                "--export",
                "--embed-svg-images",
                "--embed-diagram",
                "--svg-theme light",
            ],
        )

    name = "drawio"
    cli_name = "drawio"
    priority = 0

    @tagging.tags(click_aliases=["ls"])
    def list(self, changes=False, **kwargs):
        """
        Lists affected resources (*.drawio files) for this project
        """
        return self._list(changes=changes, **kwargs)

    @cli.click.argument("output", required=False)
    @cli.click.argument("input")
    def export(self, input, output=None, format="svg"):
        """
        Exports a given .drawio file to some
        output file/format (default format is SVG)
        """
        assert input.endswith(".drawio") or input.endswith(
            ".xml"
        ), "Expected an xml or drawio file as input"
        output = output or ".".join(
            list(filter(None, input.split(".")[:-1])) + [format]
        )
        export_args = " ".join(self.config.export_args)
        export_args += f" --width {self.config.export_width}"
        export_docker_args = "-w /workspace -v `pwd`:/workspace"
        result = self._run_docker(
            (
                f"docker run {export_docker_args} {self.config.export_docker_image} "
                f"{export_args} --output {output} {input}"
            ),
            strict=True,
        )
        print(result.stdout if result.succeeded else result)
        raise SystemExit(0 if result.succeeded else 1)

    def serve(self):
        """
        Runs the drawio-UI in a docker-container
        """
        port = self.config.http_port
        dargs = self.config.docker_args
        dimg = self.config.docker_image
        cmd_t = f"docker run {dargs} -p {port}:{port} {dimg}"
        return self._run_docker(cmd_t, strict=True, interactive=True)

    def open(self, *args, **kwargs):
        """
        Opens a browser for the container started by `serve`
        """
        return webbrowser.open(
            f"http://localhost:{self.config.http_port}/?offline=1&https=0"
        )
